package sim.math;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.List;
import java.util.StringTokenizer;

import sim.exception.SNoImplementationException;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * <p>
 * La classe <b>SVector3d</b> reprsente une vecteur <i>x</i>, <i>y</i> et <i>z</i>  trois dimensions. 
 * </p>
 * 
 * <p>
 * Les oprations mathmatiques supportes sont les suivantes : 
 * <ul>- L'addition.</ul>
 * <ul>- La soustraction.</ul>
 * <ul>- La multiplication par un scalaire.</ul>
 * <ul>- Le produit scalaire (<i>dot product</i>).</ul>
 * <ul>- Le produit vectoriel (<i>cross product</i>).</ul>
 * <ul>- La normalisation (vecteur de module 1).</ul>
 * </p>
 * 
 * @author Simon Vezina
 * @since 2014-12-16
 * @version 2017-09-20 (labo - La loi de Coulomb v1.1)
 */
public class SVector3d implements Comparable<SVector3d> {

  //---------------
  // CONSTANTES  //
  //---------------

  /**
   * La constante 'DEFAULT_COMPONENT' correspond  la composante par dfaut des variables x,y et z du vecteur tant gale  {@value}.
   */
	private static final double  DEFAULT_COMPONENT = 0.0;

	/**
	 * La constant <b>ZERO</b> correspond au vecteur nul correspondant  la coordonne (0.0, 0.0, 0.0).
	 */
	public static final SVector3d ZERO = new SVector3d(0.0, 0.0, 0.0);
	
	/**
	 * La constant <b>UNITARY_X</b> correspond au vecteur unitaire parallle  l'axe x correspondant  (1.0, 0.0, 0.0).
	 * On utilise galement la notation <b><i>i</i></b> pour le reprsenter.
	 */
	public static final SVector3d UNITARY_X = new SVector3d(1.0, 0.0, 0.0);
	
	/**
   * La constant <b>UNITARY_Y</b> correspond au vecteur unitaire parallle  l'axe y correspondant  (0.0, 1.0, 0.0).
   * On utilise galement la notation <b><i>j</i></b> pour le reprsenter.
   */
	public static final SVector3d UNITARY_Y = new SVector3d(0.0, 1.0, 0.0);
	
	/**
   * La constant <b>UNITARY_Z</b> correspond au vecteur unitaire parallle  l'axe z correspondant  (0.0, 0.0, 1.0).
   * On utilise galement la notation <b><i>k</i></b> pour le reprsenter.
   */
	public static final SVector3d UNITARY_Z = new SVector3d(0.0, 0.0, 1.0);
	
	/**
   * La constante <b>ORIGIN</b> reprsente un vecteur origine tant situ  la coordonne (0.0, 0.0, 0.0).
   * Il est gal au vecteur <b>ZERO</b>.
   */
	public static final SVector3d ORIGIN = ZERO;
	
	//--------------
	// VARIABLES  //
	//--------------
	
	/**
	 * La variable <b>x</b> correspond  la composante x du vecteur 3d.
	 */
	private final double x;
	
	/**
   * La variable <b>y</b> correspond  la composante y du vecteur 3d.
   */
	private final double y;	
	
	/**
   * La variable <b>z</b> correspond  la composante z du vecteur 3d.
   */
	private final double z;	
	
	//-----------------
	// CONSTRUCTEURS //
	//-----------------
	/**
	 * Constructeur reprsentant un vecteur 3d  l'origine d'un systme d'axe xyz.
	 */
  public SVector3d()
	{
		this(DEFAULT_COMPONENT, DEFAULT_COMPONENT, DEFAULT_COMPONENT);
	}
	
	/**
	 * Constructeur avec composante <i>x</i>,<i>y</i> et <i>z</i> du vecteur 3d.
	 * 
	 * @param x La composante <i>x</i> du vecteur.
	 * @param y La composante <i>y</i> du vecteur.
	 * @param z La composante <i>z</i> du vecteur.
	 */
	public SVector3d(double x, double y, double z)
	{
		this.x = x;
		this.y = y;
		this.z = z;
	}
	
	/**
	 * Constructeur d'un vecteur 3d en utilisant un string dcrivant les paramtres x, y et z du vecteur. 
	 * Une lecture autorise peut prendre la forme suivante : "double x" "double y" "double z" (incluant la notation avec , ( ) [ ] comme dlimieur dans l'expression du string comme par exemple (2.3, 4.3, 5.7) ).
	 * 
	 * @param string - Le string de l'expression du vecteur en paramtres x, y, et z.
	 * @throws SReadingException Si le format de la lecture n'est pas adquat.
	 */
	public SVector3d(String string)throws SReadingException
	{
		double[] tab = read(string);
		
		x = tab[0];
		y = tab[1];
		z = tab[2];
	}
	
	//------------
	// MTHODES //
	//------------
	
	/**
	 * Mthode qui donne accs  la coordonne x du vecteur.
	 * 
	 * @return La coordonne x.
	 */
	public double getX()
	{ 
	  return x;
	}
	
	/**
	 * Mthode qui donne accs  la coordonne y du vecteur.
	 * 
	 * @return La coordonne y.
	 */
	public double getY()
	{ 
	  return y;
	}
	
	/**
	 * Mthode qui donne accs  la coordonne z du vecteur.
	 * 
	 * @return La coordonne z.
	 */
	public double getZ()
	{ 
	  return z;
	}
	
	/**
	 * Mthode qui calcule <b>l'addition</b> de deux vecteurs. 
	 * 
	 * @param v Le vecteur  ajouter.
	 * @return La somme des deux vecteurs.
	 */
	public SVector3d add(SVector3d v)
	{	
		throw new SNoImplementationException("La mthode n'est pas implmente.");
	  
	}
	
	/**
	 * Mthode qui calcul la <b>soustraction</b> de deux vecteurs. 
	 * 
	 * @param v - Le vecteur  soustraire au vecteur prsent.
	 * @return La soustraction des deux vecteurs.
	 */
	public SVector3d substract(SVector3d v)
	{
	  throw new SNoImplementationException("La mthode n'est pas implmente.");
    
	}
	
	/**
	 * Mthode qui effectue la <b>multiplication d'un vecteur par une scalaire</b>.
	 * 
	 * @param m Le scalaire.
	 * @return La multiplication d'un vecteur par un scalaire.
	 */
	public SVector3d multiply(double m)
	{
	  throw new SNoImplementationException("La mthode n'est pas implmente.");
    
	}
	
	/**
	 * Mthode pour obtenir le <b>module</b> d'un vecteur.
	 * 
	 * @return Le module du vecteur.
	 */
	public double modulus()
	{
	  throw new SNoImplementationException("La mthode n'est pas implmente.");
    
	}
	
	/**
	 * Mthode pour <b>normaliser</b> un vecteur  trois dimensions. 
	 * Un vecteur normalis possde la mme orientation, mais possde une <b> longeur unitaire </b>.
	 * Si le <b>module du vecteur est nul</b>, le vecteur normalis sera le <b> vecteur nul </b> (0.0, 0.0, 0.0).
	 * 
	 * @return Le vecteur normalis.
	 * @throws SImpossibleNormalizationException Si le vecteur ne peut pas tre normalis tant trop petit (modulus() <  SMath.EPSILON) ou de longueur nulle.
	 */
	public SVector3d normalize() throws SImpossibleNormalizationException
	{
	  throw new SNoImplementationException("La mthode n'est pas implmente.");
   
	}
	
	/**
	 * Mthode pour effectuer le <b>produit scalaire</b> entre deux vecteurs.
	 * 
	 * @param v Le vecteur  mettre en produit scalaire.
	 * @return Le produit scalaire entre les deux vecteurs.
	 */
	public double dot(SVector3d v)
	{
	  throw new SNoImplementationException("La mthode n'est pas implmente.");
   
	}
	
	/**
	 * Mthode pour effectuer le <b>produit vectoriel</b> avec un autre vecteur v.
	 * <p>
	 * Cette version du produit vectoriel est implment en respectant la <b> rgle de la main droite </b>.
	 * Il est important de rappeler que le produit vectoriel est <b> anticommutatif </b> (AxB = -BxA) et que l'ordre des vecteurs doit tre adquat pour obtenir le rsultat dsir.
	 * Si l'un des deux vecteurs est <b> nul </b> ou les deux vecteurs sont <b> parallles </b>, le rsultat sera un <b> vecteur nul </b>.
	 * </p>
	 * 
	 * @param v Le second vecteur dans le produit vectoriel.
	 * @return Le rsultat du produit vectoriel.
	 */
	public SVector3d cross(SVector3d v)
	{
	  throw new SNoImplementationException("La mthode n'est pas implmente.");
    
	}

	
		
	/**
	 * Mthode pour obtenir un vecteur avec les coordonne (x,y,z) les plus petites (en considrant le signe) parmi un ensemble de vecteurs. 
	 * 
	 * @param value_list La liste des vecteurs  analyser.
	 * @return Un vecteur ayant comme coordonne les plus petites valeurs (x,y,z) trouves.
	 */
	public static SVector3d findMinValue(List<SVector3d> value_list)
  {
	  return findMinValue(value_list.toArray(new SVector3d[value_list.size()]));
  }
	
	/**
	 * Mthode pour obtenir un vecteur avec les coordonne (x,y,z) les plus petites (en considrant le signe) parmi un ensemble de vecteurs. 
	 * 
	 * @param tab - Le tableau des vecteurs  analyser.
	 * @return Un vecteur ayant comme coordonne les plus petites valeurs (x,y,z) trouves.
	 */
	public static SVector3d findMinValue(SVector3d[] tab)
	{
	  double x_min = tab[0].getX();
	  
	  for(int i = 1; i < tab.length; i++)
	    if(x_min > tab[i].getX())
	      x_min = tab[i].getX();
	  
	  double y_min = tab[0].getY();
    
    for(int i = 1; i < tab.length; i++)
      if(y_min > tab[i].getY())
        y_min = tab[i].getY();
    
    double z_min = tab[0].getZ();
    
    for(int i = 1; i < tab.length; i++)
      if(z_min > tab[i].getZ())
        z_min = tab[i].getZ();
    
    return new SVector3d(x_min, y_min, z_min);
	}
	
	/**
   * Mthode pour obtenir un vecteur avec les coordonne (x,y,z) les plus grandes (en considrant le signe) parmi un ensemble de vecteurs. 
   * 
   * @param value_list La liste des vecteurs  analyser.
   * @return Un vecteur ayant comme coordonne les plus grandes valeurs (x,y,z) trouves.
   */
  public static SVector3d findMaxValue(List<SVector3d> value_list)
  {
    return findMaxValue(value_list.toArray(new SVector3d[value_list.size()]));
  }
  
	/**
   * Mthode pour obtenir un vecteur avec les coordonne (x,y,z) les plus grandes (en considrant le signe) parmi un ensemble de vecteurs. 
   * 
   * @param tab - Le tableau des vecteurs  analyser.
   * @return Un vecteur ayant comme coordonne les plus grandes valeurs (x,y,z) trouves.
   */
  public static SVector3d findMaxValue(SVector3d[] tab)
  {
    double x_max = tab[0].getX();
    
    for(int i = 1; i < tab.length; i++)
      if(x_max < tab[i].getX())
        x_max = tab[i].getX();
    
    double y_max = tab[0].getY();
    
    for(int i = 1; i < tab.length; i++)
      if(y_max < tab[i].getY())
        y_max = tab[i].getY();
    
    double z_max = tab[0].getZ();
    
    for(int i = 1; i < tab.length; i++)
      if(z_max < tab[i].getZ())
        z_max = tab[i].getZ();
    
    return new SVector3d(x_max, y_max, z_max);
  }
  
	/**
	 * Mthode pour crire les paramtres xyz du vecteur dans un fichier en format txt. Le format de l'criture est "x"  "y"  "z" comme l'exemple suivant : 0.6  0.2  0.5
	 * 
	 * @param bw Le buffer d'criture.
	 * @throws IOException S'il y a une erreur avec le buffer d'criture.
	 * @see BufferedWriter
	 */
	public void write(BufferedWriter bw)throws IOException
	{
		bw.write(toString());
	}
	
	

	//----------------------------------------------------------------------------------------------
	//Mthodes pour calcul spcialis (afin de rduire l'allocation de mmoire lors des calculs) 
	//----------------------------------------------------------------------------------------------
 
  /**
   * Mthode pour obtenir la distance entre deux points.
   * 
   * @param A Le premier point.
   * @param B Le deuxime point.
   * @return La distance entre le point A et B.
   */
  public static double distance(SVector3d A, SVector3d B)
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
   
  }
  
  /**
   * Mthode permettant d'obtenir l'angle entre deux vecteurs A et B. L'angle sera obtenu en radian.
   * 
   * @param A Le premier vecteur.
   * @param B Le second vecteur. 
   * @return L'angle entre les deux vecteurs.
   */
  public static double angleBetween(SVector3d A, SVector3d B)
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
    
  }
  
  /**
   * Mthode pour effectuer la <b>soustraction entre deux produits scalaires</b>. Cette opration vectorielle est quivalente  l'expression
   * <ul>(A - B) . C = A . C - B . C </ul>
   * @param A - Le vecteur A de l'expression.
   * @param B - Le vecteur B de l'expression.
   * @param C - Le vecteur C de l'expression.
   * @return Le scalaire reprsentant la solution de l'quation.
   */
  public static double AdotCsubstractBdotC(SVector3d A, SVector3d B, SVector3d C)
  {
    throw new SNoImplementationException("La mthode n'est pas implmente.");
    
  }
 
 /**
  * Mthode qui effectue le <b>produit mixte</b> de trois vecteurs A, B et C. Cette opration vectorielle est quivalente  l'expression
  * <ul>(A x B) . C = (Ax*By*Cz + Bx*Cy*Az + Cx*Ay*Bz) - (Az*By*Cx + Bx*Cz*Ay + Cy*Ax*Bz)</ul>
  * Il est  remarquer que si deux des trois vecteurs sont gaux ou colinaires, le produit mixte est nul.
  * @param A - Le 1ier vecteur dans le produit mixte.
  * @param B - Le 2ime vecteur dans le produit mixte.
  * @param C - Le 3ime vecteur dans le produit mixte.
  * @return Le scalaire rsultant du produit mixte.
  */
  public static double AcrossBdotC(SVector3d A, SVector3d B, SVector3d C)
  {
    throw new SNoImplementationException("Mthode non implmente.");
  }
 
 /**
  * Mthode qui effectue le <b>triple produit vectoriel</b> de trois vecteurs A, B et C. Cette opration vectorielle est quivalente  l'expression
  * <ul>(A x B) x C = (A . C) B - (B . C) A</ul>  
  * <p>Il est important de prciser que les deux expressions
  * <ul>(A x B) x C != A x (B x C)</ul>
  * ne sont pas gaux. L'ordre de priorit des oprations doit tre bien dfini.</p>
  *  
  * @param A - Le 1ier vecteur dans le double produit vectoriel.
  * @param B - Le 2ime vecteur dans le double produit vectoriel.
  * @param C - Le 3ime vecteur dans le double produit vectoriel.
  * @return Le vecteur rsultant du produit mixte.
  */
 public static SVector3d AcrossBcrossC(SVector3d A, SVector3d B, SVector3d C)
 {
   throw new SNoImplementationException("La mthode n'est pas implmente.");
   
 }
 
 /**
  * <p>Mthode qui effectue le <b>triple produit vectoriel</b> de trois vecteurs A, B et C avec l'ordre de priorit</p>
  * <ul>D = A x (B x C)</ul>
  * <p>o D est le rsultat du triple produit vectoriel. Cette opration vectorielle est quivalente  l'expression</p>
  * <ul>A x (B x C) = (A . C) B - (A . B) C</ul>  
  * <p>et il est important de prciser que les deux expressions</p>
  * <ul>A x (B x C) != (A x B) x C</ul>
  * <p>ne sont pas gaux puisque le <b>produit vectoriel n'est pas commutatif</b>.</p>
  *  
  * @param A - Le 1ier vecteur dans le double produit vectoriel.
  * @param B - Le 2ime vecteur dans le double produit vectoriel.
  * @param C - Le 3ime vecteur dans le double produit vectoriel.
  * @return Le vecteur rsultant du triple produit vectoriel.
  */
 public static SVector3d Across_BcrossC(SVector3d A, SVector3d B, SVector3d C)
 {
   throw new SNoImplementationException("La mthode n'est pas implmente.");
   
 }  
 
 /**
  * <p>Mthode qui effectue le <b>triple produit vectoriel</b> de trois vecteurs A, B et C avec l'ordre de priorit</p>
  * <ul>D = (A x B) x C</ul>
  * <p>o D est le rsultat du triple produit vectoriel. Cette opration vectorielle est quivalente  l'expression</p>
  * <ul>(A x B) x C = ?????</ul>  
  * <p>et il est important de prciser que les deux expressions</p>
  * <ul>(A x B) x C != A x (B x C)</ul>
  * <p>ne sont pas gaux puisque le <b>produit vectoriel n'est pas commutatif</b>.</p>
  *  
  * @param A - Le 1ier vecteur dans le double produit vectoriel.
  * @param B - Le 2ime vecteur dans le double produit vectoriel.
  * @param C - Le 3ime vecteur dans le double produit vectoriel.
  * @return Le vecteur rsultant du triple produit vectoriel.
  */
 public static SVector3d AcrossB_crossC(SVector3d A, SVector3d B, SVector3d C)
 {
   throw new SNoImplementationException("Mthode non implmente.");
 }
 
 /**
  * Mthode qui effectue une opration spcialise de multiplication par un scalaire et d'addition vectorielle quivalente 
  * <ul>V = a*B + C.</ul>
  * @param a - Le scalaire  multiplier avec B.
  * @param B - Le 1ier vecteur dans l'expression vectorielle.
  * @param C - Le 2ime vecteur dans l'expression vectorielle  ajouter.
  * @return Le vecteur rsultant de cette opration spcialise.
  */
 public static SVector3d AmultiplyBaddC(double a, SVector3d B, SVector3d C)
 {
   throw new SNoImplementationException("La mthode n'est pas implmente.");
   
 }

 //------------------------
 // MTHODES UTILITAIRES //
 //------------------------
 /**
  * Mthode utilisant un string comme paramtre pour dfinir les composantes x, y et z du vecteur. 
  * Une lecture autorise peut prendre la forme suivante : "double x" "double y" "double z" 
  * (en incluant la notation avec , ( ) [ ] < > comme dlimieur dans l'expression du string comme par exemple (2.3, 4.3, 5.7) ).
  *  
  * @param string - Le string de l'expression du vecteur en paramtres x, y, et z.
  * @return un tableau de trois lments tel que x = [0], y = [1] et z = [2]. 
  * @throws SReadingException Si le format de lecture n'est pas adquat.
  */
 private double[] read(String string)throws SReadingException
 {
   StringTokenizer tokens = new StringTokenizer(string, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
   
   if(tokens.countTokens() != 3)
     throw new SReadingException("Erreur SVector3d 003 : L'expression '" + string + "' ne contient pas exactement 3 paramtres pour les composantes xyz du vecteur 3d.");
   else
   {
     String s = "";          //String  convertir en double pour les composantes x, y et z.
     String comp = "";       //Nom de la composante en lecture utilise pour l'envoie d'une Exception s'il y a lieu.
     
     try
     {
       double[] tab = new double[3]; //Tableau des 3 composantes
       
       comp = "x";
       s = tokens.nextToken();
       tab[0] = Double.valueOf(s);
       
       comp = "y";
       s = tokens.nextToken();
       tab[1] = Double.valueOf(s);
       
       comp = "z";
       s = tokens.nextToken();
       tab[2] = Double.valueOf(s);
       
       return tab;
       
     }catch(NumberFormatException e){ 
       throw new SReadingException("Erreur SVector3d 004 : L'expression '" + s +"' n'est pas valide pour dfinir la composante '" + comp + "' du vecteur en cours.");
     }
   } 
 }

 //----------------------
 // MTHODE OVERLOADED //
 //----------------------
 
 @Override
 public String toString()
 {
   return "[" + x + ", " + y + ", " + z + "]";   
 }

 @Override
 public int hashCode()
 {
   //hashcode autognr par Eclipse avec les paramtres x, y et z
   final int prime = 31;
   int result = 1;
   long temp;
   temp = Double.doubleToLongBits(x);
   result = prime * result + (int) (temp ^ (temp >>> 32));
   temp = Double.doubleToLongBits(y);
   result = prime * result + (int) (temp ^ (temp >>> 32));
   temp = Double.doubleToLongBits(z);
   result = prime * result + (int) (temp ^ (temp >>> 32));
   return result;
 }

 @Override
 public boolean equals(Object obj)
 {
   if(this == obj)
     return true;
   
   if(obj == null)
     return false;
   
   if(!(obj instanceof SVector3d))
     return false;
   
   SVector3d other = (SVector3d) obj;
   
   //Comparaison des valeurs x,y et z en utilisant la mthode de comparaison de SMath
   if(!SMath.nearlyEquals(x, other.x))
     return false;
   
   if(!SMath.nearlyEquals(y, other.y))
     return false;
   
   if(!SMath.nearlyEquals(z, other.z))
     return false;
   
   return true;
 }
 
 @Override
 public int compareTo(SVector3d o) 
 {
   // Variable dterminant la comparaison 
   // ( comp < 0 ==> plus petit )
   // ( comp > 0 ==> plus grand )
   // ( comp = 0 ==> gal )
   int comp;
   
   // Comparaison sur les x
   comp= Double.compare(this.x, o.x);
   
   if(comp != 0)
     return comp;
   
   // Comparaison sur les y
   comp= Double.compare(this.y, o.y);
   
   if(comp != 0)
     return comp;
   
   // Comparaison sur les z
   return Double.compare(this.z, o.z);  
 }
 
}//fin de la classe SVector3d
